/*
 * This is Ted Dunning's Boyer-Moore implementation with
 * modifications to handle case folding of Latin1 text and
 * handle matching 2-byte EUC chars.
 * Author: mleisher@crl.nmsu.edu (Mark Leisher)
 */
#ifndef lint
static char rcsid[] = "$Id: kwicx.c,v 1.28 1995/08/08 14:16:08 mleisher Exp $";
#endif

#include <stdio.h>
#include <limits.h>
#include <stddef.h>
#include <ctype.h>
#include <sys/types.h>
#include "kwicx.h"
#include "map_file.h"
#include "getopt.h"

#ifndef USHRT_MAX
#define USHRT_MAX 0xffff
#endif

#ifdef __GNUC__
#define INLINE inline
#else
#define INLINE
#endif

#ifdef __STDC__
extern void *malloc(size_t);
extern void *realloc(void *, size_t);
extern void free(void *);
extern char *strrchr(char *, int);
extern int atoi(char *);
extern long atol(char *);
#else
extern char *malloc();
extern char *realloc();
extern void free();
extern char *strrchr();
extern int atoi();
extern long atol();
#endif

#define ischar8(c) (((c) & 0x80) ? 1 : 0)
#define ischar16(c) (((c) & 0xff00) ? 1 : 0)
#define IsSpace(c) (!ischar16(c) && ((c) == ' ' || iscntrl(c)))

typedef enum {sensitive, insensitive} casetype;

typedef struct {
    u_char *pattern;	/* The pattern to search for.		*/
    u_long plen;	/* Real length of the pattern.		*/
    u_long *skip;	/* Skip array for this pattern.		*/
} cstr_t;

typedef struct kwicx_data {
    cstr_t *left_keep;
    cstr_t *left_discard;
    cstr_t *right_keep;
    cstr_t *right_discard;

    /*
     * Count of how many cstr_t structures are actually in use.
     */
    u_long lk_used;
    u_long ld_used;
    u_long rk_used;
    u_long rd_used;

    /*
     * Count of how many cstr_t structures have been allocated.
     */
    u_long lk_size;
    u_long ld_size;
    u_long rk_size;
    u_long rd_size;

    kwicx_match_t *matches;
    u_long matches_used;
    u_long matches_size;
} kwicx_data;

/*
 * A cache for local data.
 */
static kwicx_data localdata;

/*
 * Translation table stuff for case insensitive matching.
 */
static u_long *skip = 0;
static u_short tr[USHRT_MAX];
static u_short untr[USHRT_MAX];
static boolean trinit = false;
static casetype casemapping = sensitive;

/*
 * The name of the program if it is standalone or the default name
 * of the routine if it is being used as a library.
 */
static char *prog = "kwicx";

static kwicx_options defaults = {
    false, /* all8        */
    (u_char *) " ",   /* sepstr      */
    false, /* casefold    */
    true,  /* filenames   */
    false, /* offsets     */
    false, /* offsetstoo  */
    10,    /* preceding   */
    10,    /* following   */
    true,  /* killspace   */
    true,  /* nospace16   */
    false, /* printlower  */
    false, /* nopad       */
    (u_char *) "\n",  /* matchsep    */
    0,     /* file        */
    0L,    /* max_matches */
    0L,    /* max_misses  */
};

/*
 * The list of Latin1 characters used for case insensitive
 * comparisons.
 */
#define Agrave		0xc0
#define agrave		0xe0
#define Odia		0xd6
#define odia		0xf6
#define Ooblique	0xd8
#define Thorn		0xde
#define thorn		0xfe
#define oslash		0xf8

static int
#ifdef __STDC__
ToLower(int ch, boolean all8)
#else
ToLower(ch, all8)
int ch;
boolean all8;
#endif
{
    /*
     * Only lower the character if it is an ASCII character.
     * If it is an 8-bit character, then it is an EUC byte
     * which we don't want to change.
     */
    if (all8 == false)
        return (ch >= 'A' && ch <= 'Z') ? (ch + ('a' - 'A')) : ch;

    if (ch >= 'A' && ch <= 'Z')
      return ch + ('a' - 'A');
    else if (ch >= Agrave && ch <= Odia)
      return ch + (agrave - Agrave);
    else if (ch >= Ooblique && ch <= Thorn)
      return ch + (oslash - Ooblique);
    return ch;
}

static int
#ifdef __STDC__
ToUpper(int ch, boolean all8)
#else
ToUpper(ch, all8)
int ch;
boolean all8;
#endif
{
    /*
     * Only upper the character if it is an ASCII character.
     * If it is an 8-bit character, then it is an EUC byte
     * which we don't want to change.
     */
    if (all8 == false)
        return (ch >= 'a' && ch <= 'z') ? (ch - ('a' - 'A')) : ch;

    if (ch >= 'a' && ch <= 'z')
      return ch - ('a' - 'A');
    else if (ch >= agrave && ch <= odia)
      return ch - (agrave - Agrave);
    else if (ch >= oslash && ch <= thorn)
      return ch - (oslash - Ooblique);
    return ch;
}

static INLINE u_short
#ifdef __STDC__
nextchar(u_char *s, boolean all8, u_long end)
#else
nextchar(s, all8, end)
u_char *s;
boolean all8;
u_long end;
#endif
{
    /*
     * If we know all bytes are individual chars, or if the
     * char happens to be ASCII, just return that char.
     */
    if (all8 == true || !ischar8(*(s + end)))
      return (u_short) (*(s + end) & 0xff);

    /*
     * We know we have a 16 bit char here, so deal with it.
     */
    return (u_short) ((*(s + end) << 8) | (*(s + end + 1) & 0xff));
}

static void initskip(
#ifdef __STDC__
                     kwicx_options *opts,
                     u_char *pattern,
                     u_long plen
#endif
);

static void
#ifdef __STDC__
makepattern(kwicx_options *opts, u_char *s, cstr_t *str)
#else
makepattern(opts, s, str)
kwicx_options *opts;
u_char *s;
cstr_t *str;
#endif
{
    u_long len;
    u_short ch;
    u_char *nsp;

    if (str) {
        str->pattern = 0;
        str->plen = 0;
        str->skip = 0;
    }

    if (s == 0 || str == 0)
      return;

    len = (u_long) strlen((char *) s);
    str->pattern = nsp = (u_char *) malloc(len + 1);
    while (*s != '\0') {
        if (opts->all8)
          *nsp = (opts->casefold == false) ? *s :
            ToLower(*s, opts->all8);
        else {
            ch = nextchar(s, opts->all8, 0L);
            if (!ischar16(ch))
              *nsp = (opts->casefold == false) ? *s :
                ToLower(*s, opts->all8);
            else {
                *nsp++ = (ch >> 8) & 0xff;
                *nsp = ch & 0xff;
                s++;
            }
        }
        s++;
        nsp++;
        str->plen++;
    }
    *nsp = '\0';

    str->skip = (u_long *) malloc(sizeof(u_long) * USHRT_MAX);
    skip = str->skip;
    initskip(opts, str->pattern, str->plen);
}

static void
#ifdef __STDC__
initcase(boolean fold, boolean all8)
#else
initcase(fold, all8)
boolean fold, all8;
#endif
{
    u_short i, c;

    for (i = 0; i < USHRT_MAX; i++) {
        tr[i] = (fold == true) ? ToLower(i, all8) : i;
        if (fold == true) {
            /*
             * If i is lower then upper it.
             * If i is upper then lower it.
             */
            c = tr[i];
            untr[i] = (c != i) ? c : ToUpper(i, all8);
        } else
          untr[i] = i;
    }
    trinit = true;
}

static INLINE u_short
#ifdef __STDC__
pchar(u_char *p, u_long n, boolean all8)
#else
pchar(p, n, all8)
u_char *p;
u_long n;
boolean all8;
#endif
{
    u_long i;

    if (all8 == true)
      return (u_short) p[n];

    for (i = 0L; i < n; i++, p++) {
        if (ischar8(*p))
          p++;
    }
    return (!ischar8(*p)) ? (u_short) (*p) : (u_short) ((*p << 8) | (*(p + 1)));
}

/*
 * This function assumes that the patterns are already lowered if
 * case insensitive searching is being done.
 */
static void
#ifdef __STDC__
initskip(kwicx_options *opts, u_char *pattern, u_long plen)
#else
initskip(opts, pattern, plen)
kwicx_options *opts;
u_char *pattern;
u_long plen;
#endif
{
    u_long i;
    u_short c, ch;

    if (trinit == false)
      initcase(opts->casefold, opts->all8);

    for (i = 0L; i < USHRT_MAX; i++)
      skip[i] = plen;
    for (i = 0L; i < plen; i++) {
        ch = pchar(pattern, i, opts->all8);
        skip[ch] = plen - i - 1;
        c = untr[ch];
        while (c != ch) {
            skip[c] = plen - i - 1;
            c = untr[c];
        }
    }
}

static INLINE u_long
#ifdef __STDC__
collapse_space(kwicx_options *opts, u_char *text,
               u_long textlen, u_long offset, boolean backward)
#else
collapse_space(opts, text, textlen, offset, backward)
kwicx_options *opts;
u_char *text;
u_long textlen, offset;
boolean backward;
#endif
{
    u_char *s, *p, *end;
    boolean end16 = false;

    end = text + textlen;
    s = p = text + offset;

    if (backward == true) {
        /*
         * Go to the character following the space and see if it is
         * a 16-bit character so spaces can be compacted if need be.
         */
        while (p < end && IsSpace(*p))
          p++;
        if (p < end && opts->all8 == false && ischar8(*p))
          end16 = true;

        /*
         * Now count the number of spaces backwards.
         */
        p = s;
        while (p > text && IsSpace(*p))
          p--;

        /*
         * Determine if we are looking at a 16-bit character preceding
         * the group of spaces.
         */
        if (p > text && opts->all8 == false) {
            if (opts->nospace16 == true && end16 == true && ischar8(*p))
              return (u_long) (s - (p - 1));
        }

        /*
         * Now move up one to either point at the second byte of the EUC
         * char if it was found and the flags were set that way, or move
         * to the first space in the group of spaces.
         */
        p++;
        return (u_long) (s - p);
    } else {
        /*
         */
        while (p < end && IsSpace(*p))
          p++;
        if (opts->all8 == false) {
            if (p < end && !ischar8(*p))
              p--;
        } else
          p--;
        return (u_long) (p - s);
    }
}

static INLINE u_long
#ifdef __STDC__
fskip(u_char *text, u_long textlen, u_long start, u_long len,
      kwicx_options *opts)
#else
fskip(text, textlen, start, len, opts)
u_char *text;
u_long textlen, start, len;
kwicx_options *opts;
#endif
{
    u_char *e = text + start, *end = text + textlen;
    boolean last16 = false;

    if (e >= end)
      return textlen;
    while (len > 0 && e < end) {
        if (opts->killspace == true && IsSpace(*e)) {
            while (e < end && IsSpace(*e))
              e++;
            if (opts->nospace16 == true && last16 == true && ischar8(*e))
              e++;
            else
              e--;
        }
        last16 = false;
        if (opts->all8 == false && ischar8(*e)) {
            e++;
            if (ischar8(*e))
              last16 = true;
        }
        e++;
        len--;
    }
    return (u_long) (e - text);
}

static INLINE u_long
#ifdef __STDC__
bskip(u_char *text, u_long textlen, u_long start, u_long len,
      kwicx_options *opts)
#else
bskip(text, textlen, start, len, opts)
u_char *text;
u_long textlen, start, len;
kwicx_options *opts;
#endif
{
    u_char *e = text + start;
    boolean last16 = false;

    while (len > 0 && e > text) {
        if (opts->all8 == false && ischar8(*e)) {
            e--;
            if (ischar8(*e))
              e--;
            last16 = true;
        } else {
            e--;
            if (opts->all8 == false && ischar8(*e))
              e--;
            last16 = false;
        }
        if (opts->killspace == true && IsSpace(*e)) {
            while (e > text && IsSpace(*e))
              e--;
            if (opts->nospace16 == true && last16 == true && ischar8(*e))
              e--;
            else
              e++;
        }
        len--;
    }
    return (u_long) (e - text);
}

static boolean
#ifdef __STDC__
search(kwicx_options *opts, u_char *pattern, u_long plen,
       u_long *base, u_char *text, u_long textlen,
       u_long *start, u_long *end)
#else
search(opts, pattern, plen, base, text, textlen, start, end)
kwicx_options *opts;
u_char *pattern;
u_long plen, *base;
u_char *text;
u_long textlen, *start, *end;
#endif
{
    u_long i, j, mend, misses = 0L;
    u_short tch, pch, c;
    boolean first = true;

    if (base != 0 && *base == textlen)
      return false;

    mend = 0L;

    j = plen - 1;
    if (base != 0)
      i = fskip(text, textlen, *base, j, opts);
    else
      i = fskip(text, textlen, 0L, j, opts);
    while (i < textlen) {
        tch = nextchar(text, opts->all8, i);
        if (opts->killspace == true) {
            if (IsSpace(tch)) {
                if (first == false)
                  i -= collapse_space(opts, text, textlen, i, true);
                else
                  i += collapse_space(opts, text, textlen, i, false);
                tch = nextchar(text, opts->all8, i);
                if (IsSpace(tch))
                  tch = ' ';
            }
        }
        pch = pchar(pattern, j, opts->all8);
        c = tr[tch];
        if (c == pch) {
            if (first == true) {
                mend = i + 1 + ischar16(pch);
                first = false;
            }
            if (j == 0L) {
                *start = i;
                *end = mend;
                return true;
            }
            i = bskip(text, textlen, i, 1L, opts);
            j--;
        } else {
            if (first == false)
              first = true;
            if (plen - j > skip[c])
              i = fskip(text, textlen, i, plen - j, opts);
            else
              i = fskip(text, textlen, i, skip[c], opts);
            misses++;
            if (opts->max_misses > 0L && opts->max_misses == misses)
              break;
            j = plen - 1;
        }
    }
    if (base != 0)
      *base = i;
    return false;
}

static void
#ifdef __STDC__
getcontext(kwicx_options *opts, u_char *text, u_long textlen,
           kwicx_match_t *match)
#else
getcontext(opts, text, textlen, match)
kwicx_options *opts;
u_char *text;
u_long textlen;
kwicx_match_t *match;
#endif
{
    if (text == 0 || textlen == 0L)
      return;

    match->left_end = match->start;
    match->right_start = match->end;
    match->left_start = bskip(text, textlen, match->left_end,
                              opts->preceding, opts);
    match->right_end = fskip(text, textlen, match->right_start,
                             opts->following, opts);
    /*
     * Make sure the end of the right context is correct.
     */
    if (match->right_end > 0L && match->right_end == textlen)
      match->right_end--;
}

static void
#ifdef __STDC__
add_match(kwicx_data_t data, kwicx_match_t *match)
#else
add_match(data, match)
kwicx_data_t data;
kwicx_match_t *match;
#endif
{
    kwicx_match_t *m;
    kwicx_data_t d;

    d = (data != 0) ? data : &localdata;

    if (d->matches_used == d->matches_size) {
        if (d->matches_size == 0L)
          d->matches = (kwicx_match_t *) malloc(sizeof(kwicx_match_t));
        else
          d->matches = (kwicx_match_t *) realloc((char *) d->matches,
                                                    sizeof(kwicx_match_t) *
                                                    (d->matches_size + 1));
        d->matches_size++;
    }
    m = &d->matches[d->matches_used];
    m->file = match->file;
    m->text = match->text;
    m->textlen = match->textlen;
    m->start = match->start;
    m->end = match->end;
    m->left_start = match->left_start;
    m->right_start = match->right_start;
    m->left_end = match->left_end;
    m->right_end = match->right_end;
    d->matches_used++;
}

static boolean
#ifdef __STDC__
match_context(kwicx_options *opts, u_char *text,
              kwicx_data_t data, kwicx_match_t *match)
#else
match_context(opts, text, data, match)
kwicx_options *opts;
u_char *text;
kwicx_data_t data;
kwicx_match_t *match;
#endif
{
    u_long i, llen, rlen, ls, rs, st, en;
    cstr_t *s;
    kwicx_data_t d = (data != 0) ? data : &localdata;

    if (text == 0)
      return false;

    ls = match->left_start;
    rs = match->right_start;
    llen = match->left_end - ls;
    rlen = match->right_end - rs;

    /*
     * Set the skip array to the oskip array so we don't disturb
     * the primary skip array.
     */
    for (i = 0, s = d->left_keep; i < d->lk_used; i++, s++) {
        skip = s->skip;
        if (search(opts, s->pattern, s->plen, 0L, text+ls, llen,
                   &st, &en) == false)
          return false;
    }
      
    for (i = 0, s = d->left_discard; i < d->ld_used; i++, s++) {
        skip = s->skip;
        if (search(opts, s->pattern, s->plen, 0L, text+ls, llen,
                   &st, &en) == true)
          return false;
    }
      
    for (i = 0, s = d->right_keep; i < d->rk_used; i++, s++) {
        skip = s->skip;
        if (search(opts, s->pattern, s->plen, 0L, text+rs, rlen,
                   &st, &en) == false)
          return false;
    }
      
    for (i = 0, s = d->right_discard; i < d->rd_used; i++, s++) {
        skip = s->skip;
        if (search(opts, s->pattern, s->plen, 0L, text+rs, rlen,
                   &st, &en) == true)
          return false;
    }
    return true;
}

static void
#ifdef __STDC__
padleft(FILE *out, kwicx_options *opts, u_char *text,
        u_long start, u_long end)
#else
padleft(out, opts, text, start, end)
FILE *out;
kwicx_options *opts;
u_char *text;
u_long start, end;
#endif
{
    u_short c, lastc;
    u_char *st = text + start, *en = text + end;
    u_long count = 0;

    if (opts->all8 == true || opts->nopad == true)
      return;

    lastc = 0;
    while (st < en) {
        c = nextchar(st, opts->all8, 0L);
        if (opts->printlower)
          c = tr[c];
        if (IsSpace(c) && opts->killspace == true) {
            do {
                st++;
                c = nextchar(st, opts->all8, 0L);
            } while (IsSpace(c));
            if (opts->nospace16 == false || !ischar16(lastc) ||
                !ischar16(c))
              count++;
            lastc = ' ';
        } else {
            if (!ischar16(c))
              count++;
            lastc = c;
            st += 1 + ischar16(c);
        }
    }
    while (count > 0L) {
        putc(' ', out);
        count--;
    }
}

static void
#ifdef __STDC__
printstring(FILE *out, kwicx_options *opts, u_char *text,
            u_long start, u_long end)
#else
printstring(out, opts, text, start, end)
FILE *out;
kwicx_options *opts;
u_char *text;
u_long start, end;
#endif
{
    u_short c, lastc;
    u_char *st = text + start, *en = text + end;

    lastc = 0;
    while (st < en) {
        c = nextchar(st, opts->all8, 0L);
        if (opts->printlower)
          c = tr[c];
        if (IsSpace(c) && opts->killspace == true) {
            do {
                st++;
                c = nextchar(st, opts->all8, 0L);
            } while (st < en && IsSpace(c));
            if (opts->nospace16 == false || !ischar16(lastc) ||
                !ischar16(c))
              putc(' ', out);
            lastc = ' ';
        } else {
            if (ischar16(c))
              putc((c >> 8) & 0xff, out);
            putc(c & 0xff, out);
            lastc = c;
            st += 1 + ischar16(c);
        }
    }
}

static void
#ifdef __STDC__
printseparator(FILE *out, u_char *p)
#else
printseparator(out, p)
FILE *out;
u_char *p;
#endif
{
    if (p == 0)
      return;

    while (*p != '\0') {
        switch (*p) {
          case '\\':
            p++;
            switch (*p) {
              case 'e':
                putc(27, out);
                break;
              case 'n':
                putc('\n', out);
                break;
              case 'r':
                putc('\r', out);
                break;
              case 't':
                putc('\t', out);
                break;
              case '\\':
                putc('\\', out);
                break;
              default:
                putc(*p, out);
            }
            break;
          default:
            putc(*p, out);
        }
        p++;
    }
}

/*
 * Start public functions.
 */

kwicx_data_t
#ifdef __STDC__
kwicx_data_create(void)
#else
kwicx_data_create()
#endif
{
    kwicx_data_t data = (kwicx_data_t) malloc(sizeof(kwicx_data));

    data->lk_used = data->lk_size = 0L;
    data->ld_used = data->ld_size = 0L;
    data->rk_used = data->rk_size = 0L;
    data->rd_used = data->rd_size = 0L;
    data->matches_size = data->matches_used = 0L;
    return data;
}

void
#ifdef __STDC__
kwicx_data_reset(kwicx_data_t data)
#else
kwicx_data_reset(data)
kwicx_data_t data;
#endif
{
    u_long i, top;
    cstr_t *s;

    if (data == 0)
      return;

    top = data->lk_used;
    s = data->left_keep;
    for (i = 0L; i < top; i++, s++)
      free((char *) s->pattern);

    top = data->ld_used;
    s = data->left_discard;
    for (i = 0L; i < top; i++, s++)
      free((char *) s->pattern);

    top = data->rk_used;
    s = data->right_keep;
    for (i = 0L; i < top; i++, s++)
      free((char *) s->pattern);

    top = data->rd_used;
    s = data->right_discard;
    for (i = 0L; i < top; i++, s++)
      free((char *) s->pattern);

    data->lk_used = 0L;
    data->ld_used = 0L;
    data->rk_used = 0L;
    data->rd_used = 0L;

    data->matches_used = 0L;
}

void
#ifdef __STDC__
kwicx_data_free(kwicx_data_t data)
#else
kwicx_data_free(data)
kwicx_data_t data;
#endif
{
    u_long i, top;
    cstr_t *s;

    if (data == 0)
      return;

    top = data->lk_used;
    s = data->left_keep;
    for (i = 0L; i < top; i++, s++) {
        free((char *) s->pattern);
        free((char *) s->skip);
    }
    free((char *) data->left_keep);

    top = data->ld_used;
    s = data->left_discard;
    for (i = 0L; i < top; i++, s++) {
        free((char *) s->pattern);
        free((char *) s->skip);
    }
    free((char *) data->left_discard);

    top = data->rk_used;
    s = data->right_keep;
    for (i = 0L; i < top; i++, s++) {
        free((char *) s->pattern);
        free((char *) s->skip);
    }
    free((char *) data->right_keep);

    top = data->rd_used;
    s = data->right_discard;
    for (i = 0L; i < top; i++, s++) {
        free((char *) s->pattern);
        free((char *) s->skip);
    }
    free((char *) data->right_discard);

    if (data->matches_size > 0L)
      free((char *) data->matches);

    free((char *) data);
}

void
#ifdef __STDC__
kwicx_add_context_pattern(kwicx_options *opts, u_char *pattern,
                          kwicx_data_t data, int side, int keep)
#else
kwicx_add_context_pattern(opts, pattern, data, side, keep)
kwicx_options *opts;
u_char *pattern;
kwicx_data_t data;
int side, keep;
#endif
{
    u_long len, used;
    cstr_t *pats, *p;

    if (data == 0)
      return;

    if (side == KWICX_LEFT) {
        len = (keep == KWICX_WITH) ? data->lk_size : data->ld_size;
        used = (keep == KWICX_WITH) ? data->lk_used : data->ld_used;
        pats = (keep == KWICX_WITH) ? data->left_keep : data->left_discard;
    } else {
        len = (keep == KWICX_WITH) ? data->rk_size : data->rd_size;
        used = (keep == KWICX_WITH) ? data->rk_used : data->rd_used;
        pats = (keep == KWICX_WITH) ? data->right_keep : data->right_discard;
    }
    if (used == len) {
        if (len == 0L)
          pats = (cstr_t *) malloc(sizeof(cstr_t));
        else
          pats = (cstr_t *) realloc((char *) pats, sizeof(cstr_t) * (len + 1));
        len++;
    }
    p = &pats[used];
    makepattern(opts, pattern, p);
    used++;

    if (side == KWICX_LEFT) {
        if (keep == KWICX_WITH) {
            data->left_keep = pats;
            data->lk_used = used;
            data->lk_size = len;
        } else {
            data->left_discard = pats;
            data->ld_used = used;
            data->ld_size = len;
        }
    } else {
        if (keep == KWICX_WITH) {
            data->right_keep = pats;
            data->rk_used = used;
            data->rk_size = len;
        } else {
            data->right_discard = pats;
            data->rd_used = used;
            data->rd_size = len;
        }
    }
}

void
#ifdef __STDC__
kwicx_add_context_patterns(kwicx_options *opts, u_char **patterns,
                           u_long num_patterns, kwicx_data_t data,
                           int side, int keep)
#else
kwicx_add_context_patterns(opts, patterns, num_patterns, data, side, keep)
kwicx_options *opts;
u_char **patterns;
u_long num_patterns;
kwicx_data_t data;
int side, keep;
#endif
{
    u_long len, used, i;
    cstr_t *pats, *p;

    if (data == 0)
      return;

    if (side == KWICX_LEFT) {
        len = (keep == KWICX_WITH) ? data->lk_size : data->ld_size;
        used = (keep == KWICX_WITH) ? data->lk_used : data->ld_used;
        pats = (keep == KWICX_WITH) ? data->left_keep : data->left_discard;
    } else {
        len = (keep == KWICX_WITH) ? data->rk_size : data->rd_size;
        used = (keep == KWICX_WITH) ? data->rk_used : data->rd_used;
        pats = (keep == KWICX_WITH) ? data->right_keep : data->right_discard;
    }
    if (used == len) {
        if (len == 0L)
          pats = (cstr_t *) malloc(sizeof(cstr_t) * num_patterns);
        else
          pats = (cstr_t *) realloc((char *) pats, sizeof(cstr_t) *
                                    (len + num_patterns));
        len += num_patterns;
    }

    p = &pats[used];
    for (i = 0L; i < num_patterns; i++, p++, used++)
      makepattern(opts, patterns[i], p);

    if (side == KWICX_LEFT) {
        if (keep == KWICX_WITH) {
            data->left_keep = pats;
            data->lk_used = used;
            data->lk_size = len;
        } else {
            data->left_discard = pats;
            data->ld_used = used;
            data->ld_size = len;
        }
    } else {
        if (keep == KWICX_WITH) {
            data->right_keep = pats;
            data->rk_used = used;
            data->rk_size = len;
        } else {
            data->right_discard = pats;
            data->rd_used = used;
            data->rd_size = len;
        }
    }
}

u_long
#ifdef __STDC__
kwicx_length(kwicx_options *opts, u_char *text, u_long textlen,
             u_long start, u_long end)
#else
kwicx_length(opts, text, textlen, start, end)
kwicx_options *opts;
u_char *text;
u_long textlen, start, end;
#endif
{
    u_long tmp, cnt = 0L;
    u_char *s, *e = text + textlen;

    if (start > end) {
        tmp = start;
        start = end;
        end = tmp;
    }
    s = text + start;
    while (start < end && s < e) {
        if (opts->all8 == false && ischar8(*s)) {
            s++;
            start++;
        }
        cnt++;
        s++;
        start++;
    }
    return cnt;
}

u_long
#ifdef __STDC__
kwicx_locate(kwicx_options *opts, u_char *text, u_long textlen,
             u_long from, u_long count, char direction, u_long *length)
#else
kwicx_locate(opts, text, textlen, from, count, direction, length)
kwicx_options *opts;
u_char *text;
u_long textlen, from, count;
char direction;
u_long *length;
#endif
{
    u_char *s, *end = text + textlen;
    u_long l = 0L;

    s = text + from;
    if (direction == KWICX_BACKWARD) {
        while (count > 0L && s > text) {
            s--;
            if (opts->all8 == false && ischar8(*s))
              s--;
            count--;
            l++;
        }
    } else {
        while (count > 0L && s < end) {
            if (opts->all8 == false && ischar8(*s))
              s++;
            s++;
            count--;
            l++;
        }
    }
    *length = l;
    return (u_long) (s - text);
}

void
#ifdef __STDC__
kwicx_print_matches(FILE *out, kwicx_options *opts, kwicx_match_t *matches,
                    u_long num_matches)
#else
kwicx_print_matches(out, opts, matches, num_matches)
FILE *out;
kwicx_options *opts;
kwicx_match_t *matches;
u_long num_matches;
#endif
{
    u_long i;

    for (i = 0; i < num_matches; i++) {
        padleft(out, opts, matches[i].text, matches[i].left_start,
                matches[i].left_end);
        printstring(out, opts, matches[i].text, matches[i].left_start,
                    matches[i].left_end);
        printseparator(out, opts->sepstr);
        printstring(out, opts, matches[i].text, matches[i].start,
                    matches[i].end);
        printseparator(out, opts->sepstr);
        printstring(out, opts, matches[i].text, matches[i].right_start,
                    matches[i].right_end);
        printseparator(out, opts->matchsep);
    }
}

static unsigned long
#ifdef __STDC__
space_count(kwicx_options *opts, unsigned char *txt)
#else
space_count(opts, txt)
kwicx_options *opts;
unsigned char *txt;
#endif
{
    unsigned long cnt = 0;

    if (txt == 0 || *txt == 0)
      return 0;

    if (!IsSpace(*txt))
      return (opts->all8 == false && ischar8(*txt)) ? 2 : 1;

    while (*txt != 0 && IsSpace(*txt)) {
        cnt++;
        txt++;
    }
    return cnt;
}

boolean
#ifdef __STDC__
kwicx(FILE *out, kwicx_options *opts, u_char *pattern, u_char *text,
      u_long textlen,
      u_long *offset, kwicx_data_t data, kwicx_match_t **matches,
      u_long *num_matches)
#else
kwicx(out, opts, pattern, text, textlen, offset, data, matches, num_matches)
FILE *out;
kwicx_options *opts;
u_char *pattern, *text;
u_long textlen, *offset;
kwicx_data_t data;
kwicx_match_t **matches;
u_long *num_matches;
#endif
{
    u_long base;
    boolean gotone = false;
    kwicx_data_t dt;
    kwicx_match_t match;
    cstr_t ps;

    dt = (data != 0) ? data : &localdata;

    if (opts->max_matches > 0L && dt->matches_used == opts->max_matches) {
        if (num_matches != 0)
          *num_matches = dt->matches_used;
        if (matches != 0)
          *matches = dt->matches;
        return false;
    }
        
    if (num_matches != 0)
      *num_matches = 0L;

    if (opts->casefold != casemapping || trinit == false)
      initcase(opts->casefold, opts->all8);
    casemapping = opts->casefold;

    makepattern(opts, pattern, &ps);
    skip = ps.skip;

    base = (offset) ? *offset : 0L;
    while (search(opts, ps.pattern, ps.plen, &base,
                  text, textlen, &match.start, &match.end) == true) {
        match.file = opts->file;
        match.text = text;
        match.textlen = textlen;
        getcontext(opts, text, textlen, &match);
        if (opts->offsets == false) {
            if (match_context(opts, text, data, &match) == true) {
                gotone = true;
                if (opts->filenames == true)
                  fprintf(out, "%s: ", match.file);
                padleft(out, opts, text, match.left_start, match.left_end);
                printstring(out, opts, text, match.left_start,
                            match.left_end);
                printseparator(out, opts->sepstr);
                printstring(out, opts, text, match.start, match.end);
                printseparator(out, opts->sepstr);
                printstring(out, opts, text, match.right_start,
                            match.right_end);
                printseparator(out, opts->matchsep);
                if (opts->offsetstoo == true)
                  add_match(data, &match);
                else
                  dt->matches_used++;
            }
            skip = ps.skip;
        } else if (matches != 0 && num_matches != 0) {
            gotone = true;
            add_match(data, &match);
        }
        base = match.start + space_count(opts, text + match.start);
        if (opts->max_matches > 0L &&
            dt->matches_used == opts->max_matches)
          break;
    }
    free((char *) ps.pattern);
    free((char *) ps.skip);

    if (offset != 0)
      *offset = base;

    if (matches != 0)
      *matches = dt->matches;

    if (num_matches != 0)
      *num_matches = dt->matches_used;

    return gotone;
}

boolean
#ifdef __STDC__
kwicx_files(FILE *out, kwicx_options *opts, u_char *pattern, char *files[],
            u_short num_files, kwicx_data_t data, kwicx_match_t **matches,
            u_long *num_matches)
#else
kwicx_files(out, opts, pattern, files, num_files, data, matches, num_matches)
FILE *out;
kwicx_options *opts;
u_char *pattern;
char *files[];
u_short num_files;
kwicx_data_t data;
kwicx_match_t **matches;
u_long *num_matches;
#endif
{
    int i;
    boolean gotone = false;
    u_char *buf;
    size_t bufsize;

    for (i =0; i < num_files; i++) {
        bufsize = file_size(files[i]);
        if (bufsize != (size_t) -1) {
            buf = (u_char *) map_file(files[i], bufsize, 0);
            if (buf != (u_char *) -1) {
                opts->file = files[i];
                if (kwicx(out, opts, pattern, buf, bufsize, 0L, data,
                          matches, num_matches) == true)
                  gotone = true;
                if (unmap_file((char *) buf, bufsize) < 0) {
                    fprintf(stderr, "%s: unrecoverable error.\n", prog);
                    exit(2);
                }
            }
        }
    }

    return gotone;
}

#ifdef STANDALONE

static struct option longopts[] = {
    {"keyword",                required_argument, 0, 'k'},
    {"case-insensitive",       no_argument,       0, 'i'},
    {"left-with",              required_argument, 0, 'l'},
    {"left-without",           required_argument, 0, 'L'},
    {"right-with",             required_argument, 0, 'r'},
    {"right-wuthout",          required_argument, 0, 'R'},
    {"preceding-context-size", required_argument, 0, 'p'},
    {"following-context-size", required_argument, 0, 'f'},
    {"context-size",           required_argument, 0, 'c'},
    {"help",                   no_argument,       0, 'h'},
    {"eight-bit",              no_argument,       0, '8'},
    {"no-filenames",           no_argument,       0, 'F'},
    {"highlight-string",       required_argument, 0, 's'},
    {"no-highlight-string",    no_argument,       0, 'S'},
    {"dont-collapse-spaces",   no_argument,       0, 'C'},
    {"spaces-between-euc",     no_argument,       0, 'U'},
    {"version",                no_argument,       0, 'v'},
    {"byte-offsets",           no_argument,       0, 'b'},
    {"output-file",            required_argument, 0, 'o'},
    {"print-lower",            no_argument,       0, 'I'},
    {"print-last",             required_argument, 0, 'a'},
    {"number-matches",         required_argument, 0, 'm'},
    {0, 0, 0, 0},
};

static void
#ifdef __STDC__
usage(void)
#else
usage()
#endif
{
    fprintf(stderr, "Usage: %s [-8bhivCFISU] [-c n] [-p n] [-f n] ", prog);
    fprintf(stderr, "[-s c]\n[-l s] [-L s] [-r s] [-R s] [-o f] [-a s] -k s ");
    fprintf(stderr, "<file1> <file2> ...\n\n");

    fprintf(stderr, "-h,--help\n");
    fprintf(stderr, "\tThis help message.\n");

    fprintf(stderr, "-v,--version\n");
    fprintf(stderr, "\tPrint the version of this program.\n");

    fprintf(stderr, "-k,--keyword s\n");
    fprintf(stderr, "\tSearch for keyword `s'.\n");

    fprintf(stderr, "-i,--case-insensitive\n");
    fprintf(stderr, "\tMake search case insensitive.\n");

    fprintf(stderr, "-8,--eight-bit\n");
    fprintf(stderr, "\tAssume bytes with high bit set are not EUC chars.\n");

    fprintf(stderr, "-b,--byte-offsets\n");
    fprintf(stderr, "\tPrint a list of byte offsets instead the matches.\n");

    fprintf(stderr, "-c,--context-size n\n");
    fprintf(stderr, "\tSet preceding and following context size ");
    fprintf(stderr, "to `n' chars (default: 10).\n");

    fprintf(stderr, "-f,--following-context-size n\n");
    fprintf(stderr, "\tSet the following context size to `n' chars ");
    fprintf(stderr, "(default: 10).\n");

    fprintf(stderr, "-p,--preceding-context-size n\n");
    fprintf(stderr, "\tSet the preceding context size to `n' chars ");
    fprintf(stderr, "(default: 10).\n");

    fprintf(stderr, "-s,--highlight-string c\n");
    fprintf(stderr, "\tSet the string to print between the match ");
    fprintf(stderr, "and context (default: space).\n");

    fprintf(stderr, "-S,--no-highlight-string\n");
    fprintf(stderr, "\tDon't print ANY string between the match ");
    fprintf(stderr, "and the context.\n");

    fprintf(stderr, "-F,--no-filenames\n");
    fprintf(stderr, "\tDon't print the filename when a match occurs.\n");

    fprintf(stderr, "-I,--print-lower\n");
    fprintf(stderr, "\tPrint the match and contexts in lower case.\n");

    fprintf(stderr, "-C,--dont-collapse-spaces\n");
    fprintf(stderr, "\tDon't view multiple spaces, tabs, newlines, ");
    fprintf(stderr, "returns as single spaces.\n");

    fprintf(stderr, "-U,--spaces-between-euc\n");
    fprintf(stderr, "\tAllow collapsed spaces between EUC chars.\n");

    fprintf(stderr, "-l,--left-with s\n");
    fprintf(stderr, "\tThe pattern `s' MUST occur in the left context.\n");

    fprintf(stderr, "-L,--left-without s\n");
    fprintf(stderr, "\tThe pattern `s' MUST NOT occur in the left context.\n");

    fprintf(stderr, "-r,--right-with s\n");
    fprintf(stderr, "\tThe pattern `s' MUST occur in the right context.\n");

    fprintf(stderr, "-R,--right-without s\n");
    fprintf(stderr, "\tThe pattern `s' MUST NOT occur in the right context.\n");

    fprintf(stderr, "-o,--output-file f\n");
    fprintf(stderr, "\tThe output file to print the matches to.\n");

    fprintf(stderr, "-a,--print-last s\n");
    fprintf(stderr, "\tString to print after a match and its context are printed.\n");

    fprintf(stderr, "-m,--number-matches n\n");
    fprintf(stderr, "\tThe maximum number of matches to find.\n");

    exit(2);
}

int
#ifdef __STDC__
main(int argc, char *argv[])
#else
main(argc, argv)
int argc;
char *argv[];
#endif
{
    FILE *out = stdout;
    char *vers = rcsid;
    int opt, longind;
    boolean sval;
    u_char *keyword = 0;
    kwicx_data_t data = 0;
    kwicx_match_t *matches;
    u_long num_matches, n;
    kwicx_options opts;

    prog = strrchr(argv[0], '/');
    prog = prog ? prog + 1 : argv[0];

    opts = defaults;

    while ((opt = getopt_long(argc, argv,
                              "bhi8CFISUvk:l:L:r:R:p:f:c:s:o:a:m:",
                              longopts, &longind)) != EOF) {
        switch (opt) {
          case 'o':
            if (out != stdout)
              fclose(out);
            out = fopen(optarg, "w");
            if (out == 0) {
                fprintf(stderr, "%s: can't open output file '%s'.\n",
                        prog, optarg);
                kwicx_data_free(data);
                return 2;
            }
            break;
          case 'a':
            opts.matchsep = (u_char *) optarg;
            break;
          case 'b':
            opts.offsets = true;
            break;
          case 'h':
            usage();
            kwicx_data_free(data);
            break;
          case 'i':
            opts.casefold = true;
            break;
          case '8':
            opts.all8 = true;
            break;
          case 'v':
            vers = KWICX_VERSION;
            fprintf(stderr, "%s: %s\n", prog, vers);
            kwicx_data_free(data);
            return 2;
          case 'C':
            opts.killspace = false;
            break;
          case 'F':
            opts.filenames = false;
            break;
          case 'I':
            opts.printlower = true;
            break;
          case 'S':
            opts.sepstr = 0;
            break;
          case 'U':
            opts.nospace16 = false;
            break;
          case 'k':
            keyword = (u_char *) optarg;
            break;
          case 'p':
            opts.preceding = atoi(optarg);
            break;
          case 'f':
            opts.following = atoi(optarg);
            break;
          case 'c':
            opts.preceding = opts.following = atoi(optarg);
            break;
          case 's':
            opts.sepstr = (u_char *) optarg;
            break;
          case 'l':
            if (data == 0)
              data = kwicx_data_create();
            kwicx_add_context_pattern(&opts, (u_char *) optarg, data,
                                      KWICX_LEFT, KWICX_WITH);
            break;
          case 'L':
            if (data == 0)
              data = kwicx_data_create();
            kwicx_add_context_pattern(&opts, (u_char *) optarg, data,
                                      KWICX_LEFT, KWICX_WITHOUT);
            break;
          case 'r':
            if (data == 0)
              data = kwicx_data_create();
            kwicx_add_context_pattern(&opts, (u_char *) optarg, data,
                                      KWICX_RIGHT, KWICX_WITH);
            break;
          case 'R':
            if (data == 0)
              data = kwicx_data_create();
            kwicx_add_context_pattern(&opts, (u_char *) optarg, data,
                                      KWICX_RIGHT, KWICX_WITHOUT);
            break;
          case 'm':
            opts.max_matches = (u_long) atol(optarg);
            break;
          case '?':
          default:
            fprintf(stderr, "%s: unknown option %c.\n", prog, opt);
            usage();
        }
    }
    if (keyword == 0) {
        kwicx_data_free(data);
        fprintf(stderr, "%s: no keyword provided.\n", prog);
        return 2;
    }
    /*
     * When we get here, there should be some file names remaining
     * that are to be searched.
     */
    if (optind == argc) {
        /*
         * Nothing left on the command line so warn the user and exit.
         */
        kwicx_data_free(data);
        fprintf(stderr, "%s: no file names provided.\n", prog);
        return 2;
    }
    sval = kwicx_files(out, &opts, keyword, &argv[optind], argc - optind, data,
                       &matches, &num_matches);

    if (opts.offsets == true && num_matches > 0L) {
        for (n = 0L; n < num_matches; n++) {
            if (opts.filenames == true)
              fprintf(out, "%s: ", matches[n].file);
            fprintf(out, "%ld %ld\n", matches[n].start, matches[n].end);
        }
    }

    kwicx_data_free(data);
    if (out != stdout)
      fclose(out);

    return (sval == true) ? 0 : 1;
}

#endif /* STANDALONE */
