#ifndef _h_kwicx
#define _h_kwicx

/*
 * $Id: kwicx.h,v 1.19 1995/08/08 14:16:27 mleisher Exp $
 */

/*
 * KWICX - Key Word In Context eXtended.
 *
 * This package uses a modified version of the Boyer-Moore
 * algorithm to do searches for keywords that are ASCII, Latin1,
 * or 2-byte EUC characters.
 *
 * Author: mleisher@crl.nmsu.edu (Mark Leisher)
 */

/*
 * Version info.
 */
#define KWICX_VERSION "1.28"

/*
 * Types used in kwicx.
 */
typedef enum {false = 0, true = 1} boolean;

typedef struct {
    boolean all8;	/* Assume each eight bit byte is a charcter.
                           If this is false, then any byte with its
                           high bit set is assumed to be part of an
                           16-bit EUC character.
                           The default is false. */

    u_char *sepstr;	/* The string used to highlight the primary
                           match.  If this is 0, the primary match won't
                           be highlighted.
                           The default is ' ' (the space character).*/

    boolean casefold;	/* Do the search in a case insensitive manner.
                           The default is false. */

    boolean filenames;	/* Print the filenames at the before the match
                           and its context.
                           The default is true. */

    boolean offsets;	/* Print the byte offsets of the primary match
                           instead of the primary match and context.
                           The default is false. */

    boolean offsetstoo;	/* Generate a list of offsets as well as printing
                           the contexts. */

    u_long preceding;	/* How many "characters" long the preceding context
                           should be.
                           The default is 10. */

    u_long following;	/* How many "characters" long the following context
                           should be.
                           The default is 10. */

    boolean killspace;	/* Collapse spaces, tabs, newlines and carriage returns
                           into a single space.
                           The default is true. */

    boolean nospace16;	/* Assume no spaces occur between 16-bit EUC chars.
                           The default is true. */

    boolean printlower;	/* Print the contexts and matches in lower case. */

    boolean nopad;	/* Don't align the matches when printing. */

    u_char *matchsep;	/* String that is printed after every match. */

    char *file;		/* The filename the search is occuring in. */

    u_long max_matches; /* The maximum number of matches to make. */

    u_long max_misses;	/* The maximum number of misses to make. */
} kwicx_options;

typedef struct {
    char *file;		/* The file where the match occured.	*/
    u_char *text;	/* The text the match occured in.	*/
    u_long textlen;	/* The length of the text.		*/
    u_long start;	/* The start byte position of the match.*/
    u_long end;		/* The end byte position of the match.	*/
    u_long left_start;	/* Start of left context.		*/
    u_long left_end;	/* End of left context.			*/
    u_long right_start;	/* Start of right context.		*/
    u_long right_end;	/* End of right context.		*/
} kwicx_match_t;

/*
 * This type is used to store internal data for the kwicx routine.
 * The user does not need to do anything with this except:
 *
 * 1. Create it using kwicx_data_create() before their application
 *    calls kwicx()
 * 2. Reset it using kwicx_data_reset() whenever a new pattern is
 *    being searched for.
 * 3. Delete it using kwicx_data_free() when their application will
 *    not call kwicx() any more.
 *
 * If the caller does not have any left or right context patterns
 * to check for this data does not have to be used and a NULL
 * should be passed to the kwicx() function.
 */
typedef struct kwicx_data *kwicx_data_t;

/*
 * These are flags for adding context patterns.  Once the
 * primary expression has been matched, the user has the
 * option of making sure the left and right context contain
 * additional expressions so as to narrow their search.
 *
 * KWICX_RIGHT   - look for this expression in the context to the
 *                 right of the primary match
 * KWICX_LEFT    - look for this expression in the context to the
 *                 left of the primary match
 * KWICX_WITH    - If the expression matches, then print the match
 *                 and its context
 * KWICX_WITHOUT - If the expression matches, then *DON'T* print the match
 */
#define KWICX_RIGHT   1
#define KWICX_LEFT    2
#define KWICX_WITH    1
#define KWICX_WITHOUT 2

/*
 * These two macros are for the kwicx_locate() function.  They 
 * specify which direction to locate a byte position.
 */
#define KWICX_FORWARD  1
#define KWICX_BACKWARD 2

/*
 * Functions.
 */

extern kwicx_data_t kwicx_data_create(
#ifdef __STDC__
                                      void
#endif
);

extern void kwicx_data_reset(
#ifdef __STDC__
                             kwicx_data_t data
#endif
);

extern void kwicx_data_free(
#ifdef __STDC__
                            kwicx_data_t data
#endif
);

/*
 * An example might be:
 * kwicx_add_context_pattern("something here", data,
 *                           KWICX_LEFT, KWICX_WITH);
 *
 * This would add a pattern to be checked for on the
 * left side and if it appears then the context will
 * be printed.
 *
 * When checking for matches preceding or following the primary
 * pattern, the checking is done in a LOGICAL AND fashion.  This
 * means that first context match on either side to fail will
 * cause the primary match to be discarded.
 */
extern void kwicx_add_context_pattern(
#ifdef __STDC__
  kwicx_options *opts,	/* KWICX options record.			*/
  u_char *pattern,	/* The pattern to search for.			*/
  kwicx_data_t data,	/* The context data structure to put this in.	*/
  int side,		/* Which side to look for this pattern (left|right)*/
  int keep		/* Keep or discard the match if pattern matches	*/
#endif
);

/*
 * This function does the same thing as above except it adds a list
 * of patterns at a time instead of one.
 */
extern void kwicx_add_context_patterns(
#ifdef __STDC__
  kwicx_options *opts,
  u_char **patterns,
  u_long num_patterns,
  kwicx_data_t data,
  int side,
  int keep
#endif
);

/*
 * These macros simplify the call to kwicx_add_context_patterns().
 * The parameters are:
 *
 * kwicx_options *opts
 * u_char **patterns
 * u_long num_patterns
 * kwicx_data_t data
 *
 */
#define kwicx_add_left_with_patterns(o, p, n, d) \
  kwicx_add_context_patterns(o, p, n, d, KWICX_LEFT, KWICX_WITH)

#define kwicx_add_left_without_patterns(o, p, n, d) \
  kwicx_add_context_patterns(o, p, n, d, KWICX_LEFT, KWICX_WITHOUT)

#define kwicx_add_right_with_patterns(o, p, n, d) \
  kwicx_add_context_patterns(o, p, n, d, KWICX_RIGHT, KWICX_WITH)

#define kwicx_add_right_without_patterns(o, p, n, d) \
  kwicx_add_context_patterns(o, p, n, d, KWICX_RIGHT, KWICX_WITHOUT)

/*
 * This function determines the "real" length of part of the text.
 */
extern u_long kwicx_length(
#ifdef __STDC__
  kwicx_options *opts,
  u_char *text,
  u_long textlen,
  u_long start,
  u_long end
#endif
);

/*
 * This function takes a starting point in a buffer and scans
 * forward or backward a certain number of "real" characters.
 * What is meant by "real" is that all spaces, tabs, newlines and
 * carriage returns are counted as well as 16-bit characters (even
 * though they are two bytes long).
 * The return value is the byte offset of the location.
 */
extern u_long kwicx_locate(
#ifdef __STDC__
  kwicx_options *opts,	/* Kwicx options.				*/
  u_char *text,		/* The text to be scanned.			*/
  u_long textlen,	/* The length of the text.			*/
  u_long from,		/* The starting position to scan from.		*/
  u_long count,		/* How many "real" characters to move.		*/
  char direction,	/* The scan direction.				*/
  u_long *length	/* The "real" length of the text.		*/
#endif
);

extern void kwicx_print_matches(
#ifdef __STDC__
  FILE *out,		/* The output file.				*/
  kwicx_options *opts,	/* Kwicx options.				*/
  kwicx_match_t *matches,/* The matches array.				*/
  u_long num_matches	/* The number of matches in the match array.	*/
#endif
);

extern boolean kwicx(
#ifdef __STDC__
  FILE *out,		/* The output file for matches that print.	*/
  kwicx_options *opts,	/* Search options for kwicx.			*/
  u_char *pattern,	/* The pattern to search for.			*/
  u_char *text,		/* The text to search.				*/
  u_long textlen,	/* Length of the text to search.		*/
  u_long *offset,	/* Offset in the text to start the search.	*/
  kwicx_data_t data,	/* The context search data.			*/
  kwicx_match_t **matches,/* A list of the matches.			*/
  u_long *num_matches	/* The number of matches in the list.		*/
#endif
);

extern boolean kwicx_files(
#ifdef __STDC__
  FILE *out,		/* The output file for matches that print.	*/
  kwicx_options *opts,	/* Search options for kwicx.			*/
  u_char *pattern,	/* The pattern to search for.			*/
  char *files[],	/* The list of files in which to search.	*/
  u_short num_files,	/* Number of files in the file list.		*/
  kwicx_data_t data,	/* The context search data.			*/
  kwicx_match_t **matches,/* A list of the matches.			*/
  u_long *num_matches	/* The number of matches in the list.		*/
#endif
);

#endif /* _h_kwicx */
